#==========================================================================
#==========================================================================
#==========================================================================
#              PYTHON PACKAGE TO READ RESULTS FROM RADMC-3D
#==========================================================================
#==========================================================================
#==========================================================================

cimport cython
cimport numpy

from os.path import exists
from numpy import array, empty, linspace, fromfile, loadtxt, intc, hstack
import sys

#==========================================================================
#                        ROUTINES FOR IMAGES
#==========================================================================


#-----------------------------------------------------------------
#              READ THE RECTANGULAR TELESCOPE IMAGE
#-----------------------------------------------------------------
@cython.boundscheck(False)
def image(filename=None,ext=None,binary=False):
    
    cdef unsigned int nx=0
    cdef unsigned int ny=0
    cdef unsigned int nf=0
    cdef unsigned int index
    cdef numpy.ndarray[ndim=1, dtype=double] data
    sizepix_x=0.e0
    sizepix_y=0.e0
    
    if (filename == None):
        if (ext == None):
            if binary:
                filename = "image.bout"
            else:
                filename = "image.out"
        else:
            if binary:
                filename = "image_"+str(ext)+".bout"
            else:
                filename = "image_"+str(ext)+".out"

    if (exists(filename) == False):
        f = open("radmc3d.out","r")
        lines = f.readlines()
        f.close()
        for line in lines:
            sys.stdout.write(line)
        print("Sorry, cannot find {0:s}. Presumably radmc2d exited without success. See above for possible error messages of radmc3d!".format(filename))
        return
    else:
        if binary:
            f = open(filename, "rb")
            data = fromfile(filename)
        else:
            f = open(filename, "r")

    # Read the image.

    if binary:
        iformat = int.from_bytes(f.read(8), byteorder="little")
    else:
        iformat = int(f.readline())

    if (iformat < 1) or (iformat > 4):
        print("ERROR: File format of {0:s} not recognized.".format(filename))
        return

    if (iformat == 1) or (iformat == 3):
        radian = (1 == 0)
    else:
        radian = (1 == 1)

    if (iformat == 1) or (iformat == 2):
        stokes = (1 == 0)
    else:
        stokes = (1 == 1)

    if binary:
        nx = int.from_bytes(f.read(8), byteorder="little")
        ny = int.from_bytes(f.read(8), byteorder="little")
        nf = int.from_bytes(f.read(8), byteorder="little")
        sizepix_x, sizepix_y = tuple(data[4:6])
    else:
        nx, ny = tuple(array(f.readline().split(),dtype=int))
        nf = int(f.readline())
        sizepix_x, sizepix_y = tuple(array(f.readline().split(),dtype=float))

    if binary:
        lam = data[6:6+nf]
    else:
        lam = empty(nf)
        for i in range(nf):
            lam[i] = float(f.readline())
    
        f.readline()

    cdef numpy.ndarray[ndim=4, dtype=double] image
    if stokes:
        image = empty((ny,nx,nf,4))
    else:
        image = empty((ny,nx,nf,1))

    if binary:
        index = 6+nf
        for i in range(nf):
            for j in range(ny):
                for k in range(nx):
                    if stokes:
                        image[j,k,i,:] = data[index:index+4]
                        index += 4
                    else:
                        image[j,k,i,0] = data[index]
                        index += 1
    else:
        for i in range(nf):
            for j in range(ny):
                for k in range(nx):
                    if stokes:
                        image[j,k,i,:] = array(f.readline().split(), \
                                dtype=float)
                    else:
                        image[j,k,i,0] = float(f.readline())

                    if (j == ny-1) and (k == nx-1):
                        f.readline()

    f.close()

    # Compute the flux in this image as seen at 1 pc.

    if stokes:
        flux = image[:,:,:,0].sum(axis=0).sum(axis=0)
    else:
        flux = image.sum(axis=0).sum(axis=0)

    flux *= sizepix_x*sizepix_y

    if not radian:
        pc = 3.0857200e18
        flux /= pc**2

    # Compute the x and y coordinates

    x = linspace(-(<int>nx-1)/2.,(<int>nx-1)/2.,<int>nx)*sizepix_x
    y = linspace(-(<int>ny-1)/2.,(<int>ny-1)/2.,<int>ny)*sizepix_y
    
    return image, x, y, lam

#-----------------------------------------------------------------
#              READ AN UNSTRUCTURED IMAGE
#-----------------------------------------------------------------

def unstructured_image(filename=None,ext=None,binary=False):
    
    data = loadtxt(filename)

    x = data[:,0]
    y = data[:,1]
    
    image = data[:,4:]

    return image, x, y

#-----------------------------------------------------------------
#              READ THE CIRCULAR TELESCOPE IMAGE
#-----------------------------------------------------------------
@cython.boundscheck(False)
def circimage(filename=None,ext=None,binary=False):
    
    cdef unsigned int nx=0
    cdef unsigned int ny=0
    cdef unsigned int nf=0
    cdef unsigned int index
    cdef numpy.ndarray[ndim=1, dtype=double] data
    sizepix_x=0.e0
    sizepix_y=0.e0
    
    if (filename == None):
        if (ext == None):
            if binary:
                filename = "circimage.bout"
            else:
                filename = "circimage.out"
        else:
            if binary:
                filename = "circimage_"+str(ext)+".bout"
            else:
                filename = "circimage_"+str(ext)+".out"

    if (exists(filename) == False):
        f = open("radmc3d.out","r")
        lines = f.readlines()
        f.close()
        for line in lines:
            sys.stdout.write(line)
        print("Sorry, cannot find {0:s}. Presumably radmc2d exited without success. See above for possible error messages of radmc3d!".format(filename))
        return
    else:
        if binary:
            f = open(filename, "rb")
            data = fromfile(filename)
        else:
            f = open(filename, "r")

    # Read the image.

    if binary:
        iformat = fromfile(filename, count=1, dtype=intc).astype(int)[0]
    else:
        iformat = int(f.readline())

    if (iformat < 1) or (iformat > 4):
        print("ERROR: File format of {0:s} not recognized.".format(filename))
        return

    if (iformat == 1) or (iformat == 3):
        radian = (1 == 0)
    else:
        radian = (1 == 1)

    if (iformat == 1) or (iformat == 2):
        stokes = (1 == 0)
    else:
        stokes = (1 == 1)

    # Read in the number of pixels and frequencies.
    if binary:
        _, nr, nphi, nf = list(fromfile(filename, count=4, dtype=intc).\
                astype(int))
    else:
        nr, nphi = tuple(array(f.readline().split(),dtype=int))
        nf = int(f.readline())

    # Read in the radius information.

    if binary:
        index = 2
        re = hstack(([0.0], data[index:index+nr+1]))
        index += nr+1
        rc = hstack(([0.0], data[index:index+nr]))
        index += nr
    else:
        f.readline()
        re = empty(nr+2)
        for i in range(nr+2):
            re[i] = float(f.readline())

        f.readline()
        rc = empty(nr+1)
        for i in range(nr+1):
            rc[i] = float(f.readline())

    # Read in the phi information.

    if binary:
        phie = data[index:index+nphi+1]
        index += nphi+1
        phic = data[index:index+nphi]
        index += nphi
    else:
        f.readline()
        phie = empty(nphi+1)
        for i in range(nphi+1):
            phie[i] = float(f.readline())

        f.readline()
        phic = empty(nphi)
        for i in range(nphi):
            phic[i] = float(f.readline())

    # Read in the wavelength information.

    if binary:
        lam = data[index:index+nf]
        index += nf
    else:
        f.readline()
        lam = empty(nf)
        for i in range(nf):
            lam[i] = float(f.readline())
    
        f.readline()

    cdef numpy.ndarray[ndim=4, dtype=double] image
    if stokes:
        image = empty((nphi,nr+1,nf,4))
    else:
        image = empty((nphi,nr+1,nf,1))

    if binary:
        for i in range(nf):
            for j in range(nphi):
                for k in range(nr+1):
                    if stokes:
                        image[j,k,i,:] = data[index:index+4]
                        index += 4
                    else:
                        image[j,k,i,0] = data[index]
                        index += 1
    else:
        for i in range(nf):
            for j in range(nphi):
                for k in range(nr+1):
                    if stokes:
                        image[j,k,i,:] = array(f.readline().split(), \
                                dtype=float)
                    else:
                        image[j,k,i,0] = float(f.readline())

                    if (j == nphi-1) and (k == nr+1-1):
                        f.readline()

    f.close()

    # Compute the flux in this image as seen at 1 pc.

    if stokes:
        flux = image[:,:,:,0].sum(axis=0).sum(axis=0)
    else:
        flux = image.sum(axis=0).sum(axis=0)

    return image, rc, phic, lam

#==========================================================================
#                        ROUTINES FOR SPECTRA
#==========================================================================

#-----------------------------------------------------------------
#                       READ A SPECTRUM
#-----------------------------------------------------------------

def spectrum(filename=None,ext=None):

    if (filename == None):
        if (ext == None):
            filename = "spectrum.out"
        else:
            filename = "spectrum_"+str(ext)+".out"

    if exists(filename):
        f = open(filename, "r")
    else:
        print("Sorry, cannot find {0:s}. Presumably radmc2d exited without success. See above for possible error messages of radmc3d!".format(filename))
        return

    # Read the spectrum.

    iformat = int(f.readline())

    if (iformat != 1):
        print("ERROR: File format of {0:s} not recognized.".format(filename))
        return

    nf = int(f.readline())

    f.readline()

    lam = empty(nf)
    spectrum = empty(nf)

    for i in range(nf):
        lam[i], spectrum[i] = tuple(array(f.readline().split(),dtype=float))

    cc = 2.9979245800000e10

    freq = 1e4*cc/lam

    f.close()

    return spectrum, lam

#==========================================================================
#               ROUTINES FOR READING DENSITY AND TEMPERATURE
#==========================================================================

#--------------------------------------------------------------------------
#                    READ THE AMR GRID INFORMATION
#--------------------------------------------------------------------------

# Read in the dust temperature.

def amr_grid(filename=None, ext=None):

    f = open("amr_grid.inp","r")

    f.readline()
    f.readline()

    coordsystem = int(f.readline())
    gridinfo = int(f.readline())
    incl_x, incl_y, incl_z = tuple(f.readline().replace("\t","").\
            replace("\n","").split())
    nx, ny, nz = tuple(array(f.readline().replace("\t"," ").\
            replace("\n","").split(), dtype=int))

    x = array(f.readline().replace("\n","").split(), dtype=float)
    y = array(f.readline().replace("\n","").split(), dtype=float)
    z = array(f.readline().replace("\n","").split(), dtype=float)

    return coordsystem, gridinfo, incl_x, incl_y, incl_z, nx, ny, nz, x, y, z

# Read in the dust density.

def dust_density(filename=None, ext=None):

    f = open("dust_density.inp","r")

    f.readline()

    ncells = int(f.readline())
    nspecies = int(f.readline())

    temperature = []
    for i in range(nspecies):
        temp = empty((ncells,))

        for j in range(ncells):
            temp[j] = float(f.readline())

        temperature.append(temp)

    f.close()

    return temperature

# Read in the dust temperature.

def dust_temperature(filename=None, ext=None, binary=False):

    if (filename == None):
        if (ext == None):
            if binary:
                filename = "dust_temperature.bdat"
            else:
                filename = "dust_temperature.dat"
        else:
            if binary:
                filename = "dust_temperature_"+str(ext)+".bdat"
            else:
                filename = "dust_temperature_"+str(ext)+".dat"

    if binary:
        f = open(filename, "rb")
        data = fromfile(filename)
    else:
        f = open(filename,"r")

    if binary:
        int.from_bytes(f.read(8), byteorder="little")
        int.from_bytes(f.read(8), byteorder="little")
        ncells = int.from_bytes(f.read(8), byteorder="little")
        nspecies = int.from_bytes(f.read(8), byteorder="little")
    else:
        f.readline()
        ncells = int(f.readline())
        nspecies = int(f.readline())

    temperature = []
    index = 3
    for i in range(nspecies):
        temp = empty((ncells,))

        for j in range(ncells):
            if binary:
                temp[j] = data[index+j]
            else:
                temp[j] = float(f.readline())

        temperature.append(temp)

    f.close()

    return temperature

# Read in the scattering phase function.

def scattering_phase(filename=None, binary=False):

    if filename == None:
        if binary:
            filename = "scattering_phase.bout"
        else:
            filename = "scattering_phase.out"

    if binary:
        f = open(filename, "rb")
        data = fromfile(filename)
    else:
        f = open(filename, "r")

    f.readline()
    ncells = int(f.readline())
    nfreq = int(f.readline())

    freq = array(f.readline().split(), dtype=float)

    scattering_phase = []
    for i in range(nfreq):
        temp = empty((ncells,))

        for j in range(ncells):
            temp[j] = float(f.readline())

        scattering_phase.append(temp)

    f.close()

    return freq, scattering_phase



from sklearn.neighbors import KernelDensity
import scipy.signal
import scipy.stats
import emcee
import numpy

def load_results(config, model_path='', code="dynesty", discard=100, \
        best="median", unc="none", percentile=68, chisq_cut=False, \
        trim=None, fix_log10=False, gas_mass=False, return_format="dict"):
    r"""
    Load the results of a modeling run with pdspy.

    Args:
        :attr:`config` (module):
            The configuration module for the fit. Typically from `utils.load_config`
        :attr:`model_path` (`str`):
            The path to the directory where the information for the radiative transfer model is stored.
        :attr:`code` (`str`, optional):
            Which code was used to run the radiative transfer modeling fit. Default: `"dynesty"`
        :attr:`discard` (`int`, optional):
            If using `"emcee"`, the number of burn-in steps to discard from the beginning of the chain. Default: `100`
        :attr:`best` (str, optional):
            What method to use to determine the best-fit model parameters. Options include:
            `"median"`: Take the median of the posterior samples for each parameter.
            `"peak"`: Take the peak of the marginalized 1D posterior, as determined from a Kernel Density Estimator for each parameter.
            `"kde"`: Take the peak of the `n x d` posterior, as determined by a Kernel Density Estimator, i.e. the maximum likelihood model.
            Default: `"median"`
        :attr:`median` (`str`, optional):
            What method to use to get the uncertainty on the best fit parameters. Options include `"std"`, i.e. take the standard deviation of the samples, `"percentile"`, i.e. use the difference between `best` and supplied percentile values, or `None`, i.e. don't return uncertainties. Default: `None`
        :attr:`percentile` (`float`, optional):
            What percentile to use to calculate the uncertainties. Default: `68`
        :attr:`chisq_cut` (`bool`, optional):
            Use a chi-squared distribution to estimate from the number of parameters and samples a threshold below which samples are outliers from the distribution and trim those values. Default: `False`
        :attr:`trim` (`str`, optional):
            Can be used to supply conditions to use to cut samples from the posterior. e.g. `"loga_max < 1.0,logR_disk > 3.5"`. Parameters must be separated by a comma with no spaces. Default: `None`
        :attr:`fix_log10` (`bool`, optional):
            By default, the parameters that were fit in log-space will be returned in log-space. If `fix_log10=True`, then those values will instead be returned in linear space. Default: `False`
        :attr:`gas_mass` (`bool`, optional):
            By default, due to its history as a dust radiative transfer modeling code, `pdspy` returns disk and envelope masses as dust masses. If you would like to get the estimated gas mass, assuming a gas-to-dust ratio of 100, use `gas_mass=True`. Default: `False`
        :attr:`return_format` (`str`, optional):
            Typically, this function will return the best-fit parameters and uncertainties as a dictionary, with each value given the parameter name as a key, e.g. `{"logR_disk":1.54}`. It can also, however, return them as lists if `return_format="list"`. Default: `"dict"`


    Returns:
        :attr:`keys` (`list` of `str`):
            A list of the free parameter names from the model.
        :attr:`params` (`dict` or `list`):
            The best-fit parameter values as either a dictionary or list, depending on the value of `return_format`.
        :attr:`sigma` (`dict` or `list`, optional):
            The uncertainties if using the standard deviation. Only returned if `unc='std'"`
        :attr:`sigma_up` (`dict` or `list`, optional):
            The upper uncertainties, if using percentiles. Only returned if `unc="percentile"`
        :attr:`sigma_down` (`dict` or `list`, optional):
            The lower uncertainties, if using percentiles. Only returned if `unc="percentile"`
        :attr:`samples` (numpy.ndarray, `n x d`):
            The posterior samples from the fit.
    """

    # Get the list of parameters.

    keys = []
    for key in sorted(config.parameters.keys()):
        if not config.parameters[key]["fixed"]:
            keys.append(key)
    ndim = len(keys)

    # Load in the samples from the model fit.

    if code == "dynesty":
        samples = numpy.load(model_path+"samples.npy")
    elif code == "emcee":
        chain = numpy.load(model_path+"chain.npy")
        prob = numpy.load(model_path+"prob.npy")

        samples = chain[:,discard:,:].reshape((-1, ndim))
        prob = prob[:,discard:].reshape((-1,))
    elif code == "emcee3":
        backend = emcee.backends.HDFBackend("results.hdf5")

        samples = backend.get_chain(discard=discard, flat=True)
        prob = backend.get_log_prob(discard=discard, flat=True)

    # Make adjustments to the samples.

    if chisq_cut:
        good = 2 * (prob - prob.max()) > -scipy.stats.chi2.isf(q=0.01/\
                samples.shape[0], df=ndim)
    else:
        good = numpy.ones(samples.shape[0], dtype=bool)

    if trim != None:
        for command in trim:
            command = command.split(" ")

            for i, key in enumerate(keys):
                if key == command[0]:
                    if command[1] == '<':
                        good = numpy.logical_and(good, samples[:,i] > \
                                float(command[2]))
                    else:
                        good = numpy.logical_and(good, samples[:,i] < \
                                float(command[2]))

    samples = samples[good,:]

    # Get the best fit parameters.

    if best == "median":
        params = numpy.median(samples, axis=0)
    elif best == "peak":
        zscored_samples = (samples - samples.min(axis=0)) / \
            (samples.max(axis=0) - samples.min(axis=0))

        params = []
        for i in range(samples.shape[1]):
            bw = (samples.shape[0] * (1+2) / 4.)**(-1./(1+4)) / 3.
            kde = KernelDensity(kernel="gaussian", bandwidth=bw).\
                    fit(zscored_samples[:,i:i+1])

            xtest = numpy.linspace(samples[:,i].min(), samples[:,i].max(), 100)
            zscored_xtest = (xtest - samples[:,i].min()) / \
                    (samples[:,i].max() - samples[:,i].min())

            scores = kde.score_samples(zscored_xtest[:,None])

            params.append(xtest[numpy.argmax(scores)])

        params = numpy.array(params)
    elif best == "kde":
        zscored_samples = (samples - samples.min(axis=0)) / \
            (samples.max(axis=0) - samples.min(axis=0))

        bw = (samples.shape[0] * (ndim+2) / 4.)**(-1./(ndim+4))

        kde = KernelDensity(kernel='gaussian', bandwidth=bw).fit(\
                zscored_samples)
        scores = kde.score_samples(zscored_samples)

        params = samples[numpy.argmax(scores)]

    # If we ask for the gas mass instead of the dust mass, fix.

    if gas_mass:
        for i in range(ndim):
            if keys[i] in ["logM_disk","logM_env"]:
                params[i] += 2.
                samples[:,i] += 2.

    # Fix the parameters that were done in log10.

    if fix_log10:
        for i in range(ndim):
            if "log" in keys[i]:
                params[i] = 10.**params[i]
                samples[:,i] = 10.**samples[:,i]

    # Get the uncertainties.

    if unc == "std":
        sigma = samples.std(axis=0)
    elif unc == "percentile":
        percentiles = numpy.percentile(samples, [50-percentile/2, \
                50+percentile/2], axis=0)

        sigma_up = percentiles[1] - params
        sigma_down = params - percentiles[0]

    # Make a dictionary of the best fit parameters.

    if return_format == "dict":
        params = dict(zip(keys, params))

        if unc == "std":
            sigma = dict(zip(keys, sigma))
        elif unc == "percentile":
            sigma_up = dict(zip(keys, sigma_up))
            sigma_down = dict(zip(keys, sigma_down))

    # Finally, return the values.

    if unc == "std":
        return keys, params, sigma, samples
    elif unc == "percentile":
        return keys, params, sigma_up, sigma_down, samples
    else:
        return keys, params, samples



from matplotlib.widgets import Slider, Button, RadioButtons
import matplotlib.pyplot as plt

"""
Display a 3d ndarray with a slider to move along the third dimension.

Extra keyword arguments are passed to imshow

Original from here: http://nbarbey.github.io/2011/07/08/matplotlib-slider.html
"""

def cubeshow(cube, axis=2, **kwargs):
    # check dim
    if not cube.ndim == 3:
        raise ValueError("cube should be an ndarray with ndim == 3")

    # Generate a figure with a set of axes.

    fig = plt.figure()

    ax = plt.subplot(111)

    fig.subplots_adjust(bottom=0.17, top=0.98)

    # Pick out just the first image to show.

    s = [slice(0, 1) if i == axis else slice(None) for i in range(3)]
    im = cube[tuple(s)].squeeze()

    # Show the image.

    l = ax.imshow(im, **kwargs)

    # Create the slider

    axslide = fig.add_axes([0.3, 0.02, 0.4, 0.075])

    slider = Slider(axslide, '', 0, cube.shape[axis] - 1, valinit=0, \
            valfmt=' %i', valstep=1)

    def update(val):
        ind = int(slider.val)
        s = [slice(ind, ind + 1) if i == axis else slice(None)
                 for i in range(3)]
        im = cube[tuple(s)].squeeze()
        l.set_data(im)
        fig.canvas.draw()

    slider.on_changed(update)

    # Create a button to go to the next frame.

    axnext = fig.add_axes([0.81, 0.02, 0.1, 0.075])

    bnext = Button(axnext, 'Next')

    def next_image(event):
        if slider.val < cube.shape[axis]-1:
            slider.set_val(slider.val+1)
        elif slider.val == cube.shape[axis]-1:
            slider.set_val(0)

    bnext.on_clicked(next_image)

    # Create a button to go to the previous frame.

    axprev = fig.add_axes([0.09, 0.02, 0.1, 0.075])

    bprev = Button(axprev, 'Prev.')

    def prev_image(event):
        if slider.val > 0:
            slider.set_val(slider.val-1)
        elif slider.val == 0:
            slider.set_val(cube.shape[axis]-1)

    bprev.on_clicked(prev_image)

    # Show the plot.

    plt.show(block=True)



from numpy import arange,array,where
from numpy import abs as absv
from .line_flux import line_flux
import matplotlib.pyplot as plt

def find_lines(data,thresh=3):
    
    nside = 8
    nleft = 8
    nright = 8
    
    index = 8
    
    B = 1.0/1200
    
    while index < data.wave.size-8:
        
        lines = array([data.wave[index]])
        
        results,chisq = line_flux(data,lines,nleft=nleft,nright=nright, \
            quiet=True)
        results_fw,chisq_fw = line_flux(data,lines,nleft=nleft,nright=nright,\
            quiet=True,fixed_width=True)
        
        chisq_test = (chisq <= thresh) & (chisq_fw <= thresh)
        SN_test = (results[0,1]/results[0,2] >= 5) & \
                  (results_fw[0,1]/results_fw[0,2] >= 5)
        width_test = (results[0,3] >= lines[0]*B/2) & (results[0,3] <= \
            lines[0]*3*B)
        flux_test = absv(results[0,1]-results_fw[0,1])/results[0,1] <= 0.2
        
        #if (chisq_test & flux_test) & (SN_test & width_test):
        if (chisq_test & SN_test):
            print("    {0:f}   {1:e}   {2:e}   {3:f}   {4:f}".format( \
                results[0,0],results[0,1],results[0,2],results[0,3],chisq))
            
            results,chisq = line_flux(data,lines,nleft=nleft,nright=nright, \
                plotout="{0:6.3f}.pdf".format(results[0,0]),quiet=True)
            
            index = where(absv(data.wave-data.wave[index]-0.05) == \
                absv(data.wave-data.wave[index]-0.05).min())[0][0]
            nside = 6
        else:
            index = where(absv(data.wave-data.wave[index]-0.05) == \
                absv(data.wave-data.wave[index]-0.05).min())[0][0]
#        else:
#            if (nleft > 3) & (nright == nside):
#                nleft = nleft - 1
#            else:
#                nleft = nside
#                if (nright > 3) & (nleft == nside):
#                    nright = nright-1
#                else:
#                    nright = nside
#                    if nside > 3:
#                        nside = nside - 1
#                        nleft = nside
#                        nright = nside
#                    else:
#                        nside = 8
#                        nleft = nside
#                        nright = nside
#                        index = where(absv(data.wave-data.wave[index]-0.05) ==\
#                            absv(data.wave-data.wave[index]-0.05).min())[0][0]



import scipy.optimize
import numpy

def leastsq(func, x0, args=(), limits=None, Dfun=None, full_output=0, \
        col_deriv=0, ftol=1.49012e-08, xtol=1.49012e-08, gtol=0.0, maxfev=0, \
        epsfcn=None, factor=100, diag=None):

    def within_bounds(p):
        if (type(limits) != type(None)):
            for i in range(len(p)):
                if ((limits[i]["limited"][0] == True) and \
                        (p[i] < limits[i]["limits"][0])):
                    return False
                elif ((limits[i]["limited"][1] == True) and \
                        (p[i] > limits[i]["limits"][1])):
                    return False

            return True
        else:
            return True

    def residuals(p):
        if within_bounds(p):
            return func(p, *args)
        else:
            return numpy.repeat(1e20, len(func(p, *args)))

    return scipy.optimize.leastsq(residuals, x0, args=(), \
            full_output=full_output, col_deriv=col_deriv, ftol=ftol, \
            xtol=xtol, gtol=gtol, maxfev=maxfev, epsfcn=epsfcn, factor=factor, \
            diag=diag)
